@@ -68,7 +68,10 @@ class BaseCondition:
             conditions on multi-output networks should consider using a ``neurodiffeq.conditions.EnsembleCondition``.
         """
 
-        warnings.warn(f"`{self.__class__.__name__}.set_impose_on` is deprecated and will be removed in the future")
+        warnings.warn(
+            f"`{self.__class__.__name__}.set_impose_on` is deprecated and will be removed in the future",
+            DeprecationWarning,
+        )
         self.ith_unit = ith_unit
 
 
@@ -117,11 +120,18 @@ class EnsembleCondition(BaseCondition):
         r"""Re-parameterizes each column in output_tensor individually, using its corresponding sub-condition.
         This is useful when solving differential equations with a single, multi-output network.
 
-        :param output_tensor: Output of the neural network. Number of units (.shape[1]) must equal number of sub-conditions.
+        :param output_tensor:
+            Output of the neural network.
+            Number of units (.shape[1]) must equal number of sub-conditions.
         :type output_tensor: `torch.Tensor`
-        :param input_tensors: Inputs to the neural network; i.e., sampled coordinates; i.e., independent variables.
+        :param input_tensors:
+            Inputs to the neural network;
+            i.e., sampled coordinates;
+            i.e., independent variables.
         :type input_tensors: `torch.Tensor`
-        :return: The column-wise re-parameterized network output, concatenated across columns so that it's still one tensor.
+        :return:
+            The column-wise re-parameterized network output,
+            concatenated across columns so that it's still one tensor.
         :rtype: `torch.Tensor`
         """
         if output_tensor.shape[1] != len(self.conditions):
@@ -162,7 +172,10 @@ class IVP(BaseCondition):
     :type t_0: float
     :param u_0: The initial value of :math:`u`. :math:`u(t_0)=u_0`.
     :type u_0: float
-    :param u_0_prime: The initial derivative of :math:`u` w.r.t. :math:`t`. :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0} = u_0'`, defaults to None.
+    :param u_0_prime:
+        The initial derivative of :math:`u` w.r.t. :math:`t`.
+        :math:`\displaystyle\frac{\partial u}{\partial t}\bigg|_{t = t_0} = u_0'`.
+        Defaults to None.
     :type u_0_prime: float, optional
     """
 
@@ -290,9 +303,9 @@ class DirichletBVP2D(BaseCondition):
 
         :param output_tensor: Output of the neural network.
         :type output_tensor: `torch.Tensor`
-        :param x: :math:`x`-coordinates of inputs to the neural network; i.e., the sampled :math:`x`-coordinates
+        :param x: :math:`x`-coordinates of inputs to the neural network; i.e., the sampled :math:`x`-coordinates.
         :type x: `torch.Tensor`
-        :param y: :math:`y`-coordinates of inputs to the neural network; i.e., the sampled :math:`y`-coordinates
+        :param y: :math:`y`-coordinates of inputs to the neural network; i.e., the sampled :math:`y`-coordinates.
         :type y: `torch.Tensor`
         :return: The re-parameterized output of the network.
         :rtype: `torch.Tensor`
@@ -511,6 +524,176 @@ class IBVP1D(BaseCondition):
         )
 
 
+class DoubleEndedBVP1D(BaseCondition):
+      
+    r"""A boundary condition on a 1-D range where :math:`x\in[x_0, x_1]`.
+    The conditions should have the following parts:
+
+    - :math:`u(x_0)=u_0` or :math:`u'_x(x_0)=u'_0`,
+    - :math:`u(x_1)=u_1` or :math:`u'_x(x_1)=u'_1`,
+
+    where :math:`\displaystyle u'_x=\frac{\partial u}{\partial x}`.
+
+    :param x_min: The lower bound of x, the :math:`x_0`.
+    :type x_min: float
+    :param x_max: The upper bound of x, the :math:`x_1`.
+    :type x_max: float
+    :param x_min_val: The Dirichlet boundary condition when :math:`x = x_0`, the :math:`u(x_0)`, defaults to None.
+    :type x_min_val: callable, optional
+    :param x_min_prime: The Neumann boundary condition when :math:`x = x_0`, the :math:`u'_x(x_0)`, defaults to None.
+    :type x_min_prime: callable, optional
+    :param x_max_val: The Dirichlet boundary condition when :math:`x = x_1`, the :math:`u(x_1)`, defaults to None.
+    :type x_max_val: callable, optional
+    :param x_max_prime: The Neumann boundary condition when :math:`x = x_1`, the :math:`u'_x(x_1)`, defaults to None.
+    :type x_max_prime: callable, optional
+    :raises NotImplementedError: When unimplemented boundary conditions are configured.
+
+    .. note::
+        This condition cannot be passed to ``neurodiffeq.conditions.EnsembleCondition`` unless both boundaries uses
+        Dirichlet conditions (by specifying only ``x_min_val`` and ``x_max_val``) and ``force`` is set to True in
+        EnsembleCondition's constructor.
+    """
+    def __init__(
+            self, x_min, x_max,
+            x_min_val=None, x_min_prime=None,
+            x_max_val=None, x_max_prime=None,
+    ):
+        super().__init__()
+        n_conditions = sum(c is not None for c in [x_min_val, x_min_prime, x_max_val, x_max_prime])
+        if n_conditions != 2 or (x_min_val and x_min_prime) or (x_max_val and x_max_prime):
+            raise NotImplementedError('Sorry, this boundary condition is not implemented.')
+        self.x_min, self.x_min_val, self.x_min_prime = x_min, x_min_val, x_min_prime
+        self.x_max, self.x_max_val, self.x_max_prime = x_max, x_max_val, x_max_prime
+
+    def enforce(self, net, x):
+        r"""Enforces this condition on a network with inputs `x`
+
+        :param net: The network whose output is to be re-parameterized.
+        :type net: `torch.nn.Module`
+        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
+        :type x: `torch.Tensor`
+        :return: The re-parameterized output, where the condition is automatically satisfied.
+        :rtype: `torch.Tensor`
+
+        .. note::
+            This method overrides the default method of ``neurodiffeq.conditions.BaseCondition`` .
+            In general, you should avoid overriding ``enforce`` when implementing custom boundary conditions.
+        """
+
+        def ANN(x):
+            out = net(torch.cat([x], dim=1))
+            if self.ith_unit is not None:
+                out = out[:, self.ith_unit].view(-1, 1)
+            return out
+        ux = ANN(x)
+        if self.x_min_val is not None and self.x_max_val is not None:
+            return self.parameterize(ux, x)
+        elif self.x_min_val is not None and self.x_max_prime is not None:
+            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
+            ux1 = ANN(x1)
+            return self.parameterize(ux, x, ux1, x1)
+        elif self.x_min_prime is not None and self.x_max_val is not None:
+            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
+            ux0 = ANN(x0)
+            return self.parameterize(ux, x, ux0, x0)
+        elif self.x_min_prime is not None and self.x_max_prime is not None:
+            x0 = self.x_min * torch.ones_like(x, requires_grad=True)
+            x1 = self.x_max * torch.ones_like(x, requires_grad=True)
+            ux0 = ANN(x0)
+            ux1 = ANN(x1)
+            return self.parameterize(ux, x, ux0, x0, ux1, x1)
+        else:
+            raise NotImplementedError('Sorry, this boundary condition is not implemented.')
+    
+    def parameterize(self, u, x, *additional_tensors):
+        r"""Re-parameterizes outputs such that the boundary conditions are satisfied.
+
+        There are four boundary conditions that are
+        currently implemented:
+
+        - For Dirichlet-Dirichlet boundary condition :math:`u(x_0)=u_0` and :math:`u(x_1)=u_1`:
+
+          The re-parameterization is
+          :math:`\displaystyle u(x)=A+\tilde{x}\big(1-\tilde{x}\big)\mathrm{ANN}(x)`,
+          where :math:`\displaystyle A=\big(1-\tilde{x}\big)u_0+\big(\tilde{x}\big)u_1`.
+
+        - For Dirichlet-Neumann boundary condition :math:`u(x_0)=u_0` and :math:`u'_x(x_1)=u'_1`:
+
+          The re-parameterization is
+          :math:`\displaystyle u(x)=A(x)+\tilde{x}\Big(\mathrm{ANN}(x)-\mathrm{ANN}(x_1)+x_0 - \big(x_1-x_0\big)\mathrm{ANN}'_x(x_1)\Big)`,
+          where :math:`\displaystyle A(x)=\big(1-\tilde{x}\big)u_0+\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)u'_1`.
+
+        - For Neumann-Dirichlet boundary condition :math:`u'_x(x_0)=u'_0` and :math:`u(x_1)=u_1`:
+
+          The re-parameterization is
+          :math:`\displaystyle u(x)=A(x)+\big(1-\tilde{x}\big)\Big(\mathrm{ANN}(x)-\mathrm{ANN}(x_0)+x_1 + \big(x_1-x_0\big)\mathrm{ANN}'_x(x_0)\Big)`,
+          where :math:`\displaystyle A(x)=\tilde{x}u_1-\frac{1}{2}\big(1-\tilde{x}\big)^2\big(x_1-x_0\big)u'_0`.
+
+        - For Neumann-Neumann boundary condition :math:`u'_x(x_0)=u'_0` and :math:`u'_x(x_1)=u'_1`:
+
+          The re-parameterization is
+          :math:`\displaystyle u(x)=A(x)+\frac{1}{2}\tilde{x}^2\big(\mathrm{ANN}(x)-\mathrm{ANN}(x_1)-\frac{1}{2}\mathrm{ANN}'_x(x_1)\big(x_1-x_0\big)\big),
+          +\frac{1}{2}\big(1-\tilde{x}\big)^2\big(\mathrm{ANN}(x)-\mathrm{ANN}(x_0)+\frac{1}{2}\mathrm{ANN}'_x(x_0)\big(x_1-x_0\big)\big)`,
+          where :math:`\displaystyle A(x)=\frac{1}{2}\tilde{x}^2\big(x_1-x_0\big)u'_1 - \frac{1}{2}\big(1-\tilde{x}\big)^2\big(x_1-x_0\big)u'_0`.
+
+
+        Notations:
+
+        - :math:`\displaystyle\tilde{x}=\frac{x-x_0}{x_1-x_0}`,
+        - :math:`\displaystyle\mathrm{ANN}` is the neural network,
+        - and :math:`\displaystyle\mathrm{ANN}'_x=\frac{\partial ANN}{\partial x}`.
+
+        :param output_tensor: Output of the neural network.
+        :type output_tensor: `torch.Tensor`
+        :param x: The :math:`x`-coordinates of the samples; i.e., the spatial coordinates.
+        :type x: `torch.Tensor`
+        :param additional_tensors: additional tensors that will be passed by ``enforce``
+        :type additional_tensors: `torch.Tensor`
+        :return: The re-parameterized output of the network.
+        :rtype: `torch.Tensor`
+        """
+
+        x_tilde = (x - self.x_min) / (self.x_max - self.x_min)
+        if self.x_min_val is not None and self.x_max_val is not None:
+            return self._parameterize_dd(u, x, x_tilde)
+        elif self.x_min_val is not None and self.x_max_prime is not None:
+            return self._parameterize_dn(u, x, x_tilde, *additional_tensors)
+        elif self.x_min_prime is not None and self.x_max_val is not None:
+            return self._parameterize_nd(u, x, x_tilde, *additional_tensors)
+        elif self.x_min_prime is not None and self.x_max_prime is not None:
+            return self._parameterize_nn(u, x, x_tilde, *additional_tensors)
+        else:
+            raise NotImplementedError('Sorry, this boundary condition is not implemented.')
+    
+    
+    # When we have Dirichlet boundary conditions on both ends of the domain:
+    def _parameterize_dd(self, ux, x, x_tilde):
+        Ax = self.x_min_val * (1-x_tilde) + self.x_max_val * (x_tilde)
+        return Ax + x_tilde * (1 - x_tilde) * ux
+    
+    # When we have Dirichlet boundary condition on the left end of the domain
+    # and Neumann boundary condition on the right end of the domain:
+    def _parameterize_dn(self, ux, x, x_tilde, ux1, x1):
+        Ax = (1 - x_tilde) * self.x_min_val + 0.5 * x_tilde ** 2 * self.x_max_prime * (self.x_max - self.x_min)
+        return Ax + x_tilde * (ux - ux1 + self.x_min_val - diff(ux1, x1) * (self.x_max - self.x_min))
+        #Ax = self.x_min_val + (x - self.x_min) * self.x_max_prime
+        #return Ax + x_tilde * (ux - (self.x_max - self.x_min) * diff(ux1, x1) - ux1)
+    
+    # When we have Neumann boundary condition on the left end of the domain
+    # and Dirichlet boundary condition on the right end of the domain:
+    def _parameterize_nd(self, ux, x, x_tilde, ux0, x0):
+        Ax = x_tilde * self.x_max_val - 0.5 * (1 - x_tilde) ** 2 * self.x_min_prime * (self.x_max - self.x_min)
+        return Ax + (1 - x_tilde) * (ux - ux0 + self.x_max_val + diff(ux0, x0) * (self.x_max - self.x_min))
+        #Ax = self.x_max_val + (x - self.x_max) * self.x_min_prime
+        #return Ax + (1 - x_tilde) * (ux + (self.x_max - self.x_min) * diff(ux0, x0) - ux0)
+    
+    # When we have Neumann boundary conditions on both ends of the domain:
+    def _parameterize_nn(self, ux, x, x_tilde, ux0, x0, ux1, x1):
+        Ax = - 0.5 * (1 - x_tilde) ** 2 * (self.x_max - self.x_min) * self.x_min_prime + 0.5 * x_tilde ** 2 * (self.x_max - self.x_min) * self.x_max_prime
+        return Ax + 0.5 * x_tilde  ** 2 * (ux - ux1 - 0.5 * diff(ux1, x1)*(self.x_max - self.x_min)) + \
+            0.5 * (1 - x_tilde) ** 2 * (ux - ux0 + 0.5 * diff(ux0, x0)*(self.x_max - self.x_min))
+
+
 # TODO: reduce duplication
 class DirichletBVPSpherical(BaseCondition):
     r"""The Dirichlet boundary condition for the interior and exterior boundary of the sphere,
@@ -519,13 +702,22 @@ class DirichletBVPSpherical(BaseCondition):
     - :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`
     - :math:`u(r_1,\theta,\phi)=g(\theta,\phi)`
 
-    :param r_0: The radius of the interior boundary. When :math:`r_0 = 0`, the interior boundary collapses to a single point (center of the ball).
+    :param r_0:
+        The radius of the interior boundary.
+        When :math:`r_0 = 0`, the interior boundary collapses to a single point (center of the ball).
     :type r_0: float
-    :param f: The value of :math:`u` on the interior boundary. :math:`u(r_0, \theta, \phi)=f(\theta, \phi)`.
+    :param f:
+        The value of :math:`u` on the interior boundary.
+        :math:`u(r_0, \theta, \phi)=f(\theta, \phi)`.
     :type f: callable
-    :param r_1: The radius of the exterior boundary; if set to None, `g` must also be None.
+    :param r_1:
+        The radius of the exterior boundary.
+        If set to None, `g` must also be None.
     :type r_1: float or None
-    :param g: The value of :math:`u` on the exterior boundary. :math:`u(r_1, \theta, \phi)=g(\theta, \phi)`. If set to None, `r_1` must also be set to None.
+    :param g:
+        The value of :math:`u` on the exterior boundary.
+        :math:`u(r_1, \theta, \phi)=g(\theta, \phi)`.
+        If set to None, `r_1` must also be set to None.
     :type g: callable or None
     """
 
@@ -582,14 +774,22 @@ class InfDirichletBVPSpherical(BaseCondition):
     - :math:`\displaystyle u(r_0,\theta,\phi)=f(\theta,\phi)`,
     - :math:`\lim_{r\to+\infty}u(r,\theta,\phi)=g(\theta,\phi)`.
 
-    :param r_0: The radius of the interior boundary. When :math:`r_0=0`, the interior boundary collapses to a single point (center of the ball).
+    :param r_0:
+        The radius of the interior boundary.
+        When :math:`r_0=0`, the interior boundary collapses to a single point (center of the ball).
     :type r_0: float
-    :param f: The value of :math:`u` on the interior boundary. :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`.
+    :param f:
+        The value of :math:`u` on the interior boundary.
+        :math:`u(r_0,\theta,\phi)=f(\theta,\phi)`.
     :type f: callable
-    :param g: The value of :math:`u` at infinity. :math:`\lim_{r\to+\infty}u(r,\theta,\phi)=g(\theta,\phi)`.
+    :param g:
+        The value of :math:`u` at infinity.
+        :math:`\lim_{r\to+\infty}u(r,\theta,\phi)=g(\theta,\phi)`.
     :type g: callable
-    :param order: The smallest :math:`k` such that :math:`\lim_{r\to+\infty}u(r,\theta,\phi)e^{-kr}=0`, defaults to 1.
-    :type order: int or float, optional
+    :param order:
+        The smallest :math:`k` such that :math:`\lim_{r\to+\infty}u(r,\theta,\phi)e^{-kr}=0`.
+        Defaults to 1.
+    :type order: int or float
     """
 
     def __init__(self, r_0, f, g, order=1):
@@ -643,15 +843,24 @@ class DirichletBVPSphericalBasis(BaseCondition):
     The boundary conditions are: :math:`\mathbf{R}(r_0)=\mathbf{R}_0` and :math:`\mathbf{R}(r_1)=\mathbf{R}_1`,
     where :math:`\mathbf{R}` is a vector whose components are :math:`\big\{R_k\big\}_{k=1}^{K}`
 
-    :param r_0: The radius of the interior boundary. When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball)
+    :param r_0:
+        The radius of the interior boundary.
+        When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball).
     :type r_0: float
-    :param R_0: The value of harmonic coefficients :math:`\mathbf{R}` on the interior boundary. :math:`\mathbf{R}(r_0)=\mathbf{R}_0`.
+    :param R_0:
+        The value of harmonic coefficients :math:`\mathbf{R}` on the interior boundary.
+        :math:`\mathbf{R}(r_0)=\mathbf{R}_0`.
     :type R_0: `torch.Tensor`
-    :param r_1: The radius of the exterior boundary; if set to None, `R_1` must also be None
+    :param r_1:
+        The radius of the exterior boundary.
+        If set to None, `R_1` must also be None
     :type r_1: float or None
-    :param R_1: The value of harmonic coefficients :math:`\mathbf{R}` on the exterior boundary. :math:`\mathbf{R}(r_1)=\mathbf{R}_1`.
+    :param R_1:
+        The value of harmonic coefficients :math:`\mathbf{R}` on the exterior boundary.
+        :math:`\mathbf{R}(r_1)=\mathbf{R}_1`.
     :type R_1: `torch.Tensor`
-    :param max_degree: **[DEPRECATED]** highest degree when using spherical harmonics
+    :param max_degree:
+        **[DEPRECATED]** Highest degree when using spherical harmonics.
     :type max_degree: int
     """
 
@@ -713,15 +922,24 @@ class InfDirichletBVPSphericalBasis(BaseCondition):
     :math:`\lim_{r_0\to+\infty}\mathbf{R}(r)=\mathbf{R}_1`,
     where :math:`\mathbf{R}` is a vector whose components are :math:`\big\{R_k\big\}_{k=1}^{K}`.
 
-    :param r_0: The radius of the interior boundary. When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball)
+    :param r_0:
+        The radius of the interior boundary.
+        When r_0 = 0, the interior boundary is collapsed to a single point (center of the ball).
     :type r_0: float
-    :param R_0: The value of harmonic coefficients :math:`R` on the interior boundary. :math:`R(r_0)=R_0`.
+    :param R_0:
+        The value of harmonic coefficients :math:`R` on the interior boundary.
+        :math:`R(r_0)=R_0`.
     :type R_0: `torch.Tensor`
-    :param R_inf: The value of harmonic coefficients :math:`R` at infinity. :math:`\lim_{r\to+\infty}R(r)=R_\infty`.
+    :param R_inf:
+        The value of harmonic coefficients :math:`R` at infinity.
+        :math:`\lim_{r\to+\infty}R(r)=R_\infty`.
     :type R_inf: `torch.Tensor`
-    :param order: The smallest :math:`k` that guarantees :math:`\lim_{r \to +\infty} R(r) e^{-k r} = \bf 0`, defaults to 1
-    :type order: int or float, optional
-    :param max_degree: **[DEPRECATED]** highest degree when using spherical harmonics
+    :param order:
+        The smallest :math:`k` that guarantees :math:`\lim_{r \to +\infty} R(r) e^{-k r} = \bf 0`.
+        Defaults to 1.
+    :type order: int or float
+    :param max_degree:
+        **[DEPRECATED]** Highest degree when using spherical harmonics.
     :type max_degree: int
     """
 
