@@ -5,11 +5,13 @@ import matplotlib
 
 matplotlib.use('Agg')  # use a non-GUI backend, so plots are not shown during testing
 from math import erf, sqrt
-from pytest import raises
+from pytest import raises, warns
 from neurodiffeq.neurodiffeq import safe_diff as diff
 from neurodiffeq.generators import GeneratorSpherical, Generator3D
-from neurodiffeq.pde_spherical import NoConditionSpherical, DirichletBVPSpherical, InfDirichletBVPSpherical
-from neurodiffeq.pde_spherical import DirichletBVPSphericalHarmonics, InfDirichletBVPSphericalHarmonics
+from neurodiffeq.conditions import NoCondition
+from neurodiffeq.conditions import DirichletBVPSpherical
+from neurodiffeq.conditions import InfDirichletBVPSpherical
+from neurodiffeq.conditions import DirichletBVPSphericalBasis
 from neurodiffeq.pde_spherical import solve_spherical, solve_spherical_system
 from neurodiffeq.pde_spherical import MonitorSpherical
 from neurodiffeq.pde_spherical import MonitorSphericalHarmonics
@@ -91,7 +93,7 @@ def test_inf_dirichlet_bvp_spherical():
 
 def test_train_generator_spherical():
     pde = laplacian_spherical
-    condition = NoConditionSpherical()
+    condition = NoCondition()
     train_generator = GeneratorSpherical(size=64, r_min=0., r_max=1., method='equally-spaced-noisy')
     r, th, ph = train_generator.get_examples()
     assert (0. < r.min()) and (r.max() < 1.)
@@ -145,12 +147,23 @@ def test_monitor_spherical():
         'train': list(np.random.rand(10)),
         'valid': list(np.random.rand(10)),
     }
-    monitor.check(
-        nets,
-        conditions,
-        loss_history=loss_history,
-        analytic_mse_history=analytic_mse_history,
-    )
+    history = {
+        'train_loss': list(np.random.rand(10)),
+        'valid_loss': list(np.random.rand(10)),
+        'train_foo': list(np.random.rand(10)),
+        'valid_foo': list(np.random.rand(10)),
+        'train_bar': list(np.random.rand(10)),
+        'valid_bar': list(np.random.rand(10)),
+    }
+    with warns(FutureWarning):
+        monitor.check(nets, conditions, history=history, analytic_mse_history=analytic_mse_history)
+    with warns(FutureWarning):
+        monitor.check(nets, conditions, history=loss_history)
+    with warns(FutureWarning):
+        monitor.check(nets, conditions, history=loss_history, analytic_mse_history=analytic_mse_history)
+    with raises(ValueError):
+        monitor.check(nets, conditions, history={'train_foo': [], 'valid_foo': []})
+    monitor.check(nets, conditions, history=history)
 
 
 def test_solve_spherical_system():
@@ -197,7 +210,7 @@ def test_electric_potential_gaussian_charged_density():
     v_0 = (k * Q / r_0) * erf(r_0 / (np.sqrt(2) * sigma))
     v_1 = (k * Q / r_1) * erf(r_1 / (np.sqrt(2) * sigma))
 
-    def validate(solution, loss_history, analytical_mse):
+    def validate(solution):
         generator = GeneratorSpherical(512, r_min=r_0, r_max=r_1)
         rs, thetas, phis = generator.get_examples()
         us = solution(rs, thetas, phis, as_type="np")
@@ -209,7 +222,7 @@ def test_electric_potential_gaussian_charged_density():
     pde1 = lambda u, r, th, ph: laplacian_spherical(u, r, th, ph) + rho_f(r) / epsilon
     condition1 = DirichletBVPSpherical(r_0, lambda th, ph: v_0, r_1, lambda th, ph: v_1)
     monitor1 = MonitorSpherical(r_0, r_1, check_every=50)
-    solution1, loss_history1, analytic_mse1 = solve_spherical(
+    solution1, metrics_history = solve_spherical(
         pde1, condition1, r_0, r_1,
         max_epochs=2,
         return_best=True,
@@ -217,7 +230,7 @@ def test_electric_potential_gaussian_charged_density():
         monitor=monitor1,
         batch_size=64,
     )
-    validate(solution1, loss_history1, analytic_mse1)
+    validate(solution1)
 
     # solving the problem using spherical harmonics (laplcian computation is optimized)
     max_degree = 2
@@ -231,11 +244,11 @@ def test_electric_potential_gaussian_charged_density():
         sol[:, 0:1] = 2 * analytic_solution(r, th, ph)
         return sol
 
-    condition2 = DirichletBVPSphericalHarmonics(r_0=r_0, R_0=R_0, r_1=r_1, R_1=R_1, max_degree=max_degree)
+    condition2 = DirichletBVPSphericalBasis(r_0=r_0, R_0=R_0, r_1=r_1, R_1=R_1, max_degree=max_degree)
     monitor2 = MonitorSphericalHarmonics(r_0, r_1, check_every=50, max_degree=max_degree)
     net2 = FCNN(n_input_units=1, n_output_units=(max_degree + 1) ** 2)
     harmonics_fn = RealSphericalHarmonics(max_degree=max_degree)
-    solution2, loss_history2, analytic_mse2 = solve_spherical(
+    solution2, metrics_history = solve_spherical(
         pde2, condition2, r_0, r_1,
         net=net2,
         max_epochs=2,
@@ -246,7 +259,7 @@ def test_electric_potential_gaussian_charged_density():
         harmonics_fn=harmonics_fn,
     )
 
-    validate(solution2, loss_history2, analytic_mse2)
+    validate(solution2)
 
 
 def test_spherical_harmonics():
@@ -358,7 +371,7 @@ def test_spherical_harmonics():
     assert torch.max(abs_diff) <= 1e-5, f"difference too large, check again:\n {abs_diff.max()}"
 
 
-def test_spherical_laplcian():
+def test_spherical_laplacian():
     N_FLOAT = np.float64
     T_FLOAT = torch.float64
     n_samples = 10
